/************************************************************************/
/* read_o3_tab.c                                                        */
/*                                                                      */
/* Library for Stamnes table related calculations.                      */
/* Tables are generated by the Perl script Gen_o3_tab.pl.               */
/*                                                                      */
/* Author: Bernhard Mayer,                                              */
/*        Fraunhofer Institute for Atmospheric Environmental Research,  */
/*        82467 Garmisch-Partenkirchen,                                 */
/*        Germany                                                       */
/*                                                                      */
/* Modified by Arve Kylling (arve.kylling@nilu.no)                      */
/*----------------------------------------------------------------------*/
/* Copyright (C) 1997 Bernhard Mayer                                    */
/*                                                                      */
/* This program is free software; you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation; either version 1, or (at your option)  */
/* any later version.                                                   */
/*                                                                      */
/* This program is distributed in the hope that it will be useful,      */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY of FITNESS FOR A PARTICULAR PURPOSE. See the         */
/* GNU General Public License for more details.                         */
/*                                                                      */
/* To obtain a copy of the GNU General Public License write to the      */
/* Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,   */
/* USA.                                                                 */
/*----------------------------------------------------------------------*/
/************************************************************************/

/*  @12c@                

The Perl script @code{Gen_o3_tab.pl} is used to generate a matrix of
ozone values for solar zenith angle versus a chosen ratio of global
irradiance at different wavelengths. The table is read by the C program
@code{read_o3_tab} which, for a solar zenith angle and a measured ratio,
returns the overhead ozone column. All available options are displayed
when executing

@example
perl Gen_o3_tab.pl --help
@end example

@noindent and

@example
read_o3_tab -h
@end example

The following different types of tables may be generated.
    @c12@ */

/*  @12_01c@                
The simplest type of table is made of ratios of the global irradiance 
at two single wavelengths. This is the type of table described by 
Stamnes et al. (Applied Optics, {@bf 30}, 4418--4426, 1991). This type
of table is typically used to analyse measurements of the global 
irradiance from spectroradiometers.
It is generated by the following command (\ is line continuation character) 

@example
perl Gen_o3_tab.pl --slitfunction slitfncfile --lower_lambda 305. \
                   --upper_lambda 340. --file table.dat
@end example

Here @file{slitfncfile} is the name of the slit function file. It is
a two column file where the first column is the wavelength (nm, in relative
units) and the second column holds the slit function. The slit function must
be normalized to unity at the center wavelength.

The generated table @file{table.dat} is read by @code{read_o3_tab} for a
measured ratio, @code{-r 10.0}, and solar zenith angle, @code{-s 30.0}, 
corresponding to the modelled ratio in the table

@example
read_o3_tab -r 10.0 -s 30.0 table.dat
@end example

    @c12_01@ */

/*  @12_02c@                
Instead of using single wavelengths it may be of advantage to use ratios
of irradiances covering a certain wavelength range and weighted with
a bandpass function. 
This approach may reduce problems due to changes
in cloud cover and experimental uncertainties. This approach is also suitable
to calculate ozone columns from multichannel, moderate bandwidth filter
instruments (A. Dahlback, Applied Optics, {@bf 35}, 6514--6521, 1996).
Such tables are generated by

@example

perl Gen_o3_tab.pl --slitfunction slitfncfile  --lower_lambda 305.0 \ 
                   --upper_lambda 320.0 --file table.dat \ 
                   --bandpasslower bplow.dat --bandpassupper bpupp.dat
@end example

Here @file{bplow.dat} and @file{bpupp.dat} are the bandpass function of the 
lower and upper wavelength region respectively. The bandpass files have two
columns. The first column is the wavelength in nm and relative units to 
--lower_lambda and --upper_lambda. If absolute units are specified as for
filter instruments, use the --absolute option. The second column is the
bandpass function.

The tables are read in the same way as the simple wavelength ratio tables.

    @c12_02@ */

#include <stdio.h>
#include <stdlib.h>
#include <ascii.h>
#include <getopt.h>
#include <numeric.h>

typedef struct {
  int      n_zenith;
  int      n_ratio;
  double*   zenith;
  double*   ratio;
  double**  ozone;
} STAMNES_TABLE;

/* Error codes */
#define NOT_A_VALID_STAMNES_TABLE -30
#define RATIO_OUT_OF_RANGE        -31
#define ERROR_SPLINING_ZENITH     -32

#define PROGRAM "read_o3_tab"
#define VERSION "1.0"

static void print_usage (char *name)
{
  fprintf (stderr, "\nusage:   %s -r ratio -s sza [-ph] Stamne_table_file\n\n", name);
  fprintf (stderr, "%s calculates the ozone column from a measured irradiance\n", PROGRAM);
  fprintf (stderr, "ratio for a specific solar zenith angle. A Stamnes table file must\n");
  fprintf (stderr, "be present. See Gen_o3_tab.pl for information on how to generate\n");
  fprintf (stderr, "such a table\n\n");
  fprintf (stderr, "Required arguments:\n");
  fprintf (stderr, "  -r  <value>    The measured irradiance ratio corresponding to\n");
  fprintf (stderr, "                 the modelled ratio in the Stamnes table.\n");
  fprintf (stderr, "  -s  <value>    The solar zenith angle.\n");
  fprintf (stderr, "Optional arguments:\n");
  fprintf (stderr, "  -h            This help message.\n");
  fprintf (stderr, "  -p            Turn of printing of messages.\n\n");
}

/**************************************************************/
/* Read Stamnes table from file filename.                     */
/* zenith [0..n_zenith-1] is an array of solar zenith angles, */ 
/* ratio [0..n_ratio-1] an array of ratios found in the       */
/* file. The two-dimensional integer array ozone holds the    */
/* corresponding ozone columns.                               */
/* Memory for arrays is allocated automatically.              */
/**************************************************************/

int read_stamnes_table (char *filename, 
			STAMNES_TABLE **table)
{
  int i=0, j=0, status=0;
  int rows=0, columns=0, max_columns=0;
  double **value=NULL;


  /* read file to two-dimensional double array value */
  status = ASCII_file2double (filename,
			      &rows, &max_columns, &columns,
			      &value);

  if (status != 0) 
    return status;


  /* check if a rectangular matrix */  
  if (columns != max_columns)  { 
    ASCII_free_double (value, rows);
    fprintf (stderr, " ... read_stamnes_table(): %s is not a rectangular matrix!\n", filename);
    fprintf (stderr, " ... minimum number of columns: %d, maximum number of columns: %d\n", 
	     columns, max_columns);

    return NOT_A_VALID_STAMNES_TABLE;
  }
    

  /* allocate memory for Stamnes table */
  *table = calloc (1, sizeof(STAMNES_TABLE));

  
  /* first row/column contains ratio/zenith */
  (*table)->n_zenith = columns-1;
  (*table)->n_ratio  = rows-1;
 
  /* allocate memory for arrays */
  (*table)->zenith = (double *)  calloc ((*table)->n_zenith, sizeof (double));
  (*table)->ratio  = (double *)  calloc ((*table)->n_ratio,  sizeof (double));
  (*table)->ozone  = (double **) calloc ((*table)->n_ratio,  sizeof (double *));

  for (i=0; i<(*table)->n_ratio; i++)
    (*table)->ozone[i] = (double *) calloc ((*table)->n_zenith,  sizeof (double));
  
  /* copy first row to array zenith */
  for (i=1; i<=(*table)->n_zenith; i++)  
    (*table)->zenith[i-1] = value[0][i];

  /* copy first column to array ratio */
  for (j=1; j<=(*table)->n_ratio; j++)  
    (*table)->ratio[j-1] = value[j][0];



  /* copy the remainder to 2dim double array ozone */
  for (i=1; i<=(*table)->n_ratio; i++) 
    for (j=1; j<=(*table)->n_zenith; j++)
      (*table)->ozone[i-1][j-1] = value[i][j];
    

  /* free memory of double array */
  ASCII_free_double (value, rows);
 
  return 0;  /* if o.k. */
}

/**************************************************************/
/* free memory of data read from Stamnes table.               */
/**************************************************************/

void free_stamnes_table (STAMNES_TABLE *table)
{
  int i=0;

  for (i=0; i<table->n_ratio; i++) 
    free (table->ozone[i]);

  free(table->ozone);

  free(table->zenith);
  free(table->ratio);

  free(table);
}



/***************************************************************/
/* stamnes_calculate uses the data read from a Stamnes table   */
/* to calculate the ozone column for given zenith angle and    */
/* irradiance ratio.                                           */
/*                                                             */
/* zenith[0..n_zenith-1]  zenith angles found in Stamnes table */
/* ratio[0..n_ratio-1]    ratio's found in Stamnes table       */
/* ozone[0..n_ratio-1][0..n_zenith-1]   ozone columns found in */
/*   Stamnes table.                                            */
/* meas_zenith   zenith angle to be used for calculation       */
/* meas_ratio    ratio to be used for calculation              */
/* column                                                      */
/***************************************************************/

int stamnes_calculate (STAMNES_TABLE *table,
		       double meas_zenith, double meas_ratio,
		       double *column)
{
  int counter=0, status=0;
  double *a0=NULL, *a1=NULL, *a2=NULL, *a3=NULL;
  double *b0=NULL, *b1=NULL, *b2=NULL, *b3=NULL;
  double lower=0, upper=0;

  /* reset ozone column */
  *column = -1;

  /* look for neighbour points in  array of ratios */
  while (counter < table->n_ratio && table->ratio[counter] < meas_ratio)
    counter++;

  /* check if meas_ratio within Stamnes table limits */ 
  if (counter<=0 || counter >= table->n_ratio)
    return RATIO_OUT_OF_RANGE;


  /* calculate spline coefficients for ozone as function of zenith */
  status = spline_coeffc (table->zenith, table->ozone[counter-1], table->n_zenith, 
			  &a0, &a1, &a2, &a3);

  if (status!=0)  {
    free(a0);
    free(a1);
    free(a2);
    free(a3);
    return ERROR_SPLINING_ZENITH;
  }
    

  status = spline_coeffc (table->zenith, table->ozone[counter], table->n_zenith, 
			  &b0, &b1, &b2, &b3);
  
  if (status!=0)  {
    free(a0);
    free(a1);
    free(a2);
    free(a3);
    free(b0);
    free(b1);
    free(b2);
    free(b3);
    return ERROR_SPLINING_ZENITH;
  }

  /* calculate interpolated data */
  status =  calc_splined_value (meas_zenith, &lower, 
				table->zenith, table->n_zenith, 
				a0, a1, a2, a3);

  free(a0);
  free(a1);
  free(a2);
  free(a3);
  

  if (status!=0)  {
    free(b0);
    free(b1);
    free(b2);
    free(b3);
    return ERROR_SPLINING_ZENITH;
  }

  status =  calc_splined_value (meas_zenith, &upper, 
				table->zenith, table->n_zenith, 
				b0, b1, b2, b3);

  free(b0);
  free(b1);
  free(b2);
  free(b3);


  if (status!=0) 
    return ERROR_SPLINING_ZENITH;


  /* now interpolate linearly for ratio */
  *column = lower + (upper - lower) / (table->ratio[counter] - table->ratio[counter-1]) *
    (meas_ratio - table->ratio[counter-1]);

  return 0;  /* if o.k. */
}

void panic()   /* ;-) */
{
  fprintf (stderr, "Help!\n");
  exit(-1);
}

int main(int argc, char **argv) 
{
  int status=0;
  int c=0;
  int print =1, rat=0, zen=0;
  char filename[FILENAME_MAX+200] = "";
  char *dummy=NULL;

  double meas_ratio=0.0;  /* measured ratio 340 / 305 */ 
  double meas_zenith=0.0; /* solar zenith angle       */ 
  double column=0.0;
	
  STAMNES_TABLE *table = NULL;

    /* check number of remaining command line arguments */

  while ((c=getopt (argc, argv, "pr:s:?h")) != EOF)  {
    switch(c)  {
    case 'p':  /* no print messages */  
      print = 0;
      break;   
    case 'r':  /* no print messages */  
      meas_ratio  = strtod (optarg, &dummy);
      rat = 1;
      break;   
    case 's':  /* no print messages */  
      meas_zenith = strtod (optarg, &dummy);
      zen = 1;
      break;   
    case 'h':
      print_usage (argv[0]);
      return (-1);
      break;
    case '?':
      print_usage (argv[0]);
      return (-1);
      break;
    default:
      print_usage (argv[0]);
      return (-1);
    }
  }
  if (argc - optind != 1)  {
    fprintf(stderr,"No Stamnes table file specified\n"); 
    print_usage (argv[0]);
    return -1;
  }
  else if (!zen)  {
    fprintf(stderr,"No solar zenith angle given\n"); 
    print_usage (argv[0]);
    return -1;
  }
  else if (!rat)  {
    fprintf(stderr,"No measured irradiance given\n"); 
    print_usage (argv[0]);
    return -1;
  }
  
  strcpy (filename, argv[optind]);
  status = read_stamnes_table (filename, &table);
  /*  status = read_stamnes_table ("stamnes.dat", &table);*/

  if (status!=0) {
    switch (status) {
    case ASCIIFILE_NOT_FOUND:
      fprintf(stderr,"Stamnes table file not found %s\n",filename);
      break;
    default:
      fprintf(stderr,"Problem reading Stamnes table file %s\n",filename);
      break;
    }
    exit(-1);
  }

  status =  stamnes_calculate (table, meas_zenith, meas_ratio,
			       &column);

  if (status!=0)
    panic();  

  if ( print ) {
    fprintf (stdout, "the ozone column is %g DU\n", column);
  }
  fprintf (stdout, "%g\n", column);

  free_stamnes_table (table);

  return 0;
}





